#include "event_system.h"
#include "aliases.h"
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <sys/mman.h>

#define INVALID_EVENT ((Event) { .id = 0 })
#define INVALID_LISTENER ((EventListener){ .event = INVALID_EVENT, .id = 0 })

#define IS_INVALID_EVENT(EV) (EV.id == 0)
#define IS_INVALID_LISTENER(LS) (LS.id == 0)

struct event_system {
  u64  alloc_size;
  u64  event_capacity;
  u64  event_count;
  u64  free_count;
  u64  initial_callback_capacity;
  uptr events_offset;
  uptr free_offset;
  uptr callbacks_offset;
};

typedef struct event_slot EventSlot;
struct event_slot {
  uptr offset;
  bool registered;
};

typedef struct event_internal EventInternal;
struct event_internal {
  u64           capacity;
  u64           count;
  EventCallback *callbacks[];
};

void          *alloc(u64 size);
EventSlot     *get_event_slots_array(const EventSystem *es);
u64           *get_free_array(const EventSystem *es);
EventInternal *get_event(const EventSystem *es, Event event);

u32 es_init(EventSystem **es, u64 initial_event_capacity, u64 initial_listeners_capacity) {
  if (!es) {
    return ES_INIT_NULL_POINTER;
  }

  if (*es) {
    return ES_INIT_ALREADY_INITIALISED;
  }

  // Double event capacity to reduce chances of reallocation
  u64 event_capacity     = initial_event_capacity * 2;
  // Increase listener capacity by one to account for 0th EventListener being invalid
  u64 listeners_capacity = initial_listeners_capacity + 1;
  // Double listener capacity to reduce chances of reallocation
  u64 callback_capacity  = listeners_capacity * 2;
  u64 system             = sizeof(EventSystem);
  // Increase event capacity by one to account for 0th Event being invalid
  u64 event_offsets      = (event_capacity + 1) * sizeof(EventSlot);
  u64 free_slots         = event_capacity * sizeof(u64);
  u64 events             = event_capacity * sizeof(EventInternal);
  u64 callbacks          = event_capacity * callback_capacity * sizeof(EventCallback *);
  u64 alloc_size         = system + event_offsets + free_slots + events + callbacks;

  u8 *buffer = alloc(alloc_size);
  if (!buffer) {
    return ES_INIT_ALLOCATION_FAILED;
  }

  memset(buffer, 0, alloc_size);

  *es = (EventSystem *)buffer;

  EventSystem *event_system = *es;

  event_system->alloc_size                = alloc_size;
  event_system->event_capacity            = event_capacity;
  event_system->event_count               = 0;
  event_system->free_count                = 0;
  event_system->initial_callback_capacity = callback_capacity - 1;
  event_system->events_offset             = (uptr)system;
  event_system->free_offset               = (uptr)(system + event_offsets);
  event_system->callbacks_offset          = (uptr)(system + event_offsets + free_slots);

  EventSlot *ev_slots_array = get_event_slots_array(event_system);

  u64 event_item_capacity = sizeof(EventInternal) + callback_capacity * sizeof(EventCallback *);
  for (u64 i = 0; i < event_capacity; ++i) {
    ev_slots_array[i].offset = event_system->callbacks_offset + i * event_item_capacity;
  }

  return ES_INIT_SUCCESS;
}

void es_deinit(EventSystem **es) {
  if (!es || !*es) {
    return;
  }

  EventSystem *system = *es;

  munmap(system, system->alloc_size);
  *es = NULL;
}

Event es_register_event(EventSystem *es) {
  Event event = INVALID_EVENT;

  if (!es) {
    goto RETURN_EVENT;
  }

  EventSlot *event_slots_array = get_event_slots_array(es);

  if (es->free_count > 0) {
    u64 *free_array = get_free_array(es);

    event.id = free_array[--(es->free_count)];
    event_slots_array[event.id].registered = true;

    free_array[es->free_count] = 0;

    goto RETURN_EVENT;
  }

  if (es->event_count >= es->event_capacity) {
    goto RETURN_EVENT;
  }

  event.id = ++(es->event_count);
  event_slots_array[event.id].registered = true;

RETURN_EVENT:
  return event;
}

void es_deregister_event(EventSystem *es, Event event) {
  if (!es || IS_INVALID_EVENT(event)) {
    return;
  }

  EventSlot *event_slots_array = get_event_slots_array(es);
  u64 *free_array              = get_free_array(es);

  event_slots_array[event.id].registered = false;
  free_array[(es->free_count)++]         = event.id;

  if (event.id == es->event_count) {
    es->event_count -= 1;
  }
}

EventListener es_add_event_listener(EventSystem *es, u64 id, EventCallback *listener) {}

void es_remove_event_listener(EventSystem *es, EventListener listener) {}

void es_emit_event(EventSystem *es, u64 id, void *data) {}

// INTERNAL

void *alloc(u64 size) {
  return mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE | MAP_NORESERVE, -1, 0);
}

EventSlot *get_event_slots_array(const EventSystem *es) {
  return (EventSlot *)((uptr)es + es->events_offset);
}

u64 *get_free_array(const EventSystem *es) {
  return (u64 *)((uptr)es + es->free_offset);
}

EventInternal *get_event(const EventSystem *es, Event event) {
  EventSlot *event_slots_array = get_event_slots_array(es);
  return (EventInternal *)((uptr)es + event_slots_array[event.id].offset);
}
